#ifndef TCPSERVER_CC
#define TCPSERVER_CC

//
// simple tcp server class
// 
// Markus Geisler <geisler@primusnetz.de>
//

#include "TCPserver.h"

// constructor
TCPserver::TCPserver(int tmp_port, char *tmp_interface)
{
	assert(tmp_port>0 && tmp_interface!=NULL);

	int i;

	port = tmp_port;
	interface = strdup(tmp_interface);
	running = 0;
	notrunning = MAXTHREADS;

	syslog(7,"[mother] starting multithreaded tcp server version: 0.2\n");

	// initalize thread status matrix
	for(i=0;i<MAXTHREADS;i++)
	{
		// thread id
		ThreadMatrix[i].pointer = NULL;
		ThreadMatrix[i].status = NOTRUNNING;		
	}	

	myFeedbackBuffer = new FeedbackBuffer();
};

// destructor
TCPserver::~TCPserver()
{

};

int TCPserver::connect()
{
	// create socket
	mysocket = socket(AF_INET, SOCK_STREAM, 0);

	if(mysocket<0) 
	{
		perror("Can not open socket!");
		return ERROR;
	}
  
	// bind server port
	server_addr.sin_family = AF_INET;
	server_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	server_addr.sin_port = htons(port);

	if(bind(mysocket, (struct sockaddr *) &server_addr, sizeof(server_addr))<0) 
	{
		perror("Can not open port!");
		return ERROR;
	}

	// listen on socket
	listen(mysocket,5);

	return(OK);
};

char *TCPserver::wait()
{
	int tmp_mysocket;
	int n;

	// Thread id
        pthread_t tid;
	pthread_mutex_t *myMutex;

	struct in_addr source_adress;

	n = 0;
	myMutex = (pthread_mutex_t *)malloc(sizeof(pthread_mutex_t));

	// mutex initalisieren
	pthread_mutex_init(myMutex,NULL);

	// ueberpruefungs-thread erzeugen
	if(pthread_create(&tid, NULL, cyclicMatrixCheck, (void *)this ) == 0)
	{
		// ok - thread erzeugt
	}
	else
	{
	}

	while(1) 
	{
		syslog(7,"[mother] threads running: %i, threads notrunning: %i\n",running,notrunning);
		syslog(7,"[mother] waiting for data on port TCP %u\n",port);
		checkThreadMatrix();
	
		client_len = sizeof(client_addr);

		// waiting for client connection!
		tmp_mysocket = accept(mysocket, (struct sockaddr *) &client_addr, (socklen_t *) &client_len);

		if(tmp_mysocket < 0) 
		{
			perror("Connection could not be established!");
			return ERROR;
		}

		source_adress.s_addr = client_addr.sin_addr.s_addr;

		syslog(7,"[mother] connected to client '%s'!\n",inet_ntoa(source_adress));

		// build thread and pass tmp_socket to thread. listen to further connects
		nextThread = newThread(tmp_mysocket,myMutex,strdup(inet_ntoa(source_adress)));
		if(nextThread)
		{
			// ok, Thread wurde erfolgreich erzeugt
		}
		else
		{
			// Thread konnte nicht erzeugt werden. Connection closen!
			perror("No threads available!");
			close(tmp_mysocket);
		}
	}

	pthread_mutex_destroy(myMutex);
};

int TCPserver::disconnect()
{
	close(mysocket);
	return(OK);
};

// initalize new thread
TCPserverThread *TCPserver::newThread(int socket,pthread_mutex_t *mutex,char *myaddr)
{
	assert(socket > 0);

	int i;

	i = 0;

	// run through thread status matrix
	for(i=0;i<MAXTHREADS;i++)
	{
		// eintrag in matrix suchen, die noch nicht zugewiesen ist
		if(ThreadMatrix[i].pointer==NULL)
		{
			// thread id
			ThreadMatrix[i].status = RUNNING;		
			ThreadMatrix[i].pointer = new TCPserverThread;
			ThreadMatrix[i].pointer->setSocket(socket);
			ThreadMatrix[i].pointer->setMutex(mutex);
			ThreadMatrix[i].pointer->setId(i);
			ThreadMatrix[i].pointer->setFeedbackBuffer(myFeedbackBuffer);
			ThreadMatrix[i].addr = myaddr;

			// thread erzeugen
			if(pthread_create(&ThreadMatrix[i].id, NULL, ThreadMatrix[i].pointer->wrapper, (void *)ThreadMatrix[i].pointer) == 0)
			{
				// ok - thread erzeugt
			}
			else
			{
				// Thread konnte nicht erzeugt werden
				perror("Thread could not be created!");
				return ERROR;
			}

			// status variablen setzen
			running++;
			notrunning--;

			// ok - found one. bail out
			return(ThreadMatrix[i].pointer);
		}	
	}

	perror("No threads available!");
	return(ERROR);
}

// check thread matrix
void TCPserver::checkThreadMatrix()
{
	int i;

	i = 0;

	// nur berprfen, wenn auch threads laufen
	if(running)
	{

#ifdef DEBUG
		syslog(7,"[mother] checking thread matrix (%i running, %i notrunning)\n",running,notrunning);
#endif

		// run through thread status matrix
		for(i=0;i<MAXTHREADS;i++)
		{
			// nachfolgendes ueberpruefen, nur dann, wenn auch pointer da
			if(ThreadMatrix[i].pointer!=NULL)
			{
				// ueberpruefung, ob threads noch alle laufen
				if(ThreadMatrix[i].status == ThreadMatrix[i].pointer->isRunning())
				{
					// ok - Thread laeuft noch
				}
				else
				{
					// Thread-Status stimmt nicht mit reellem status ueberein
					syslog(7,"[mother] deleting thread %i. Reason: no longer running.\n",(int)ThreadMatrix[i].id);
	
					pthread_cancel(ThreadMatrix[i].id);
					delete(ThreadMatrix[i].pointer);
	
					// matrix neu setzen
					ThreadMatrix[i].status = NOTRUNNING;
					ThreadMatrix[i].id = 0;
					ThreadMatrix[i].pointer = 0;
					ThreadMatrix[i].addr = NULL;
	
					// status variablen setzen
					running--;
					notrunning++;
				}
			}
#ifdef DEBUG
			syslog(7,"[mother] child %i: Status: %i, Pointer: %i, ID: %i, Client: %s\n",i,ThreadMatrix[i].status,(int)ThreadMatrix[i].pointer,(int)ThreadMatrix[i].id,ThreadMatrix[i].addr);
#endif
		}
	}
}

// wrapper member, wegen pthread
void *TCPserver::cyclicMatrixCheck(void *object)
{
        TCPserver *myServer = (TCPserver *)object;

	while(1)
	{
        	myServer->checkThreadMatrix();
		sleep(5);
	}
}


#endif
